@@ -1,5 +1,5 @@
 /* GDK - The GIMP Drawing Kit
- * Copyright (C) 2002 Hans Breuer
+ * Copyright (C) 2002,2005 Hans Breuer
  * Copyright (C) 2003 Tor Lillqvist
  *
  * This library is free software; you can redistribute it and/or
@@ -148,7 +148,7 @@ _gdk_monitor_init (void)
 	{
 	  _gdk_monitors[i].x += _gdk_offset_x;
 	  _gdk_monitors[i].y += _gdk_offset_y;
-	  GDK_NOTE (MISC, g_print ("Monitor %d: %dx%d@+%d+%d\n",
+	  GDK_NOTE (MISC, g_print ("Monitor %d: %dx%d@%+d%+d\n",
 				   i, _gdk_monitors[i].width,
 				   _gdk_monitors[i].height,
 				   _gdk_monitors[i].x, _gdk_monitors[i].y));
@@ -175,11 +175,53 @@ _gdk_monitor_init (void)
 
 }
 
+/*
+ * Dynamic version of ProcessIdToSessionId() form Terminal Service.
+ * It is only returning something else than 0 when running under
+ * Terminal Service, available since NT4 SP4 and not for win9x
+ */
+static guint
+get_session_id (void)
+{
+  typedef BOOL (WINAPI *t_ProcessIdToSessionId) (DWORD, DWORD*);
+  static t_ProcessIdToSessionId p_ProcessIdToSessionId = NULL;
+  static HMODULE kernel32 = NULL;
+  DWORD id = 0;
+
+  if (kernel32 == NULL)
+    {
+      kernel32 = GetModuleHandle ("kernel32.dll");
+
+      g_assert (kernel32 != NULL);
+
+      p_ProcessIdToSessionId = (t_ProcessIdToSessionId) GetProcAddress (kernel32, "ProcessIdToSessionId");
+   }
+  if (p_ProcessIdToSessionId)
+      p_ProcessIdToSessionId (GetCurrentProcessId (), &id); /* got it (or not ;) */
+
+  return id;
+}
+
 GdkDisplay *
 gdk_display_open (const gchar *display_name)
 {
-  if (_gdk_display != NULL)
-    return NULL; /* single display only */
+  GDK_NOTE (MISC, g_print ("gdk_display_open: %s\n", (display_name ? display_name : "NULL")));
+
+  if (display_name == NULL ||
+      g_ascii_strcasecmp (display_name,
+			  gdk_display_get_name (_gdk_display)) == 0)
+    {
+      if (_gdk_display != NULL)
+	{
+	  GDK_NOTE (MISC, g_print ("... return _gdk_display\n"));
+	  return _gdk_display;
+	}
+    }
+  else
+    {
+      GDK_NOTE (MISC, g_print ("... return NULL\n"));
+      return NULL;
+    }
 
   _gdk_display = g_object_new (GDK_TYPE_DISPLAY, NULL);
   _gdk_screen = g_object_new (GDK_TYPE_SCREEN, NULL);
@@ -197,13 +239,63 @@ gdk_display_open (const gchar *display_name)
   g_signal_emit_by_name (gdk_display_manager_get (),
 			 "display_opened", _gdk_display);
 
+  GDK_NOTE (MISC, g_print ("... _gdk_display now set up\n"));
+
   return _gdk_display;
 }
 
 G_CONST_RETURN gchar *
 gdk_display_get_name (GdkDisplay *display)
 {
-  return gdk_get_display_arg_name ();
+  HDESK hdesk = GetThreadDesktop (GetCurrentThreadId ());
+  char dummy;
+  char *desktop_name;
+  HWINSTA hwinsta = GetProcessWindowStation ();
+  char *window_station_name;
+  DWORD n;
+  DWORD session_id;
+  char *display_name;
+  const char *retval;
+
+  n = 0;
+  GetUserObjectInformation (hdesk, UOI_NAME, &dummy, 0, &n);
+  if (n == 0)
+    desktop_name = "Default";
+  else
+    {
+      n++;
+      desktop_name = g_alloca (n + 1);
+      memset (desktop_name, 0, n + 1);
+
+      if (!GetUserObjectInformation (hdesk, UOI_NAME, desktop_name, n, &n))
+	desktop_name = "Default";
+    }
+
+  n = 0;
+  GetUserObjectInformation (hwinsta, UOI_NAME, &dummy, 0, &n);
+  if (n == 0)
+    window_station_name = "WinSta0";
+  else
+    {
+      n++;
+      window_station_name = g_alloca (n + 1);
+      memset (window_station_name, 0, n + 1);
+
+      if (!GetUserObjectInformation (hwinsta, UOI_NAME, window_station_name, n, &n))
+	window_station_name = "WinSta0";
+    }
+
+  display_name = g_strdup_printf ("%ld\\%s\\%s",
+				  get_session_id (), window_station_name,
+				  desktop_name);
+
+  retval = g_quark_to_string (g_quark_from_string (display_name));
+
+  g_free (display_name);
+
+  GDK_NOTE (MISC, g_print ("gdk_display_get_name: %s\n", retval));
+
+  return retval;
 }
 
 gint
