@@ -96,6 +96,8 @@ struct _GtkIconThemePrivate
   GList *dir_mtimes;
 
   gulong reset_styles_idle;
+
+  gboolean check_reload;
 };
 
 struct _GtkIconInfo
@@ -208,10 +210,10 @@ static void         do_theme_change   (GtkIconTheme     *icon_theme);
 
 static void  blow_themes               (GtkIconTheme    *icon_themes);
 
-static void  icon_data_free            (GtkIconData          *icon_data);
-static void load_icon_data (IconThemeDir *dir,
-			    const char   *path,
-			    const char   *name);
+static void  icon_data_free            (GtkIconData     *icon_data);
+static void load_icon_data             (IconThemeDir    *dir,
+			                const char      *path,
+			                const char      *name);
 
 static IconSuffix theme_dir_get_icon_suffix (IconThemeDir *dir,
 					     const gchar  *icon_name,
@@ -1128,7 +1130,7 @@ ensure_valid_themes (GtkIconTheme *icon_theme)
 {
   GtkIconThemePrivate *priv = icon_theme->priv;
   GTimeVal tv;
-  
+
   if (priv->themes_valid)
     {
       g_get_current_time (&tv);
@@ -1138,7 +1140,26 @@ ensure_valid_themes (GtkIconTheme *icon_theme)
     }
   
   if (!priv->themes_valid)
-    load_themes (icon_theme);
+    {
+      load_themes (icon_theme);
+      
+      if (!priv->check_reload)
+	{	  
+	  static GdkAtom atom_iconthemes = GDK_NONE;
+	  GdkEvent *event = gdk_event_new (GDK_CLIENT_EVENT);
+	  int i;
+
+	  if (!atom_iconthemes)
+	    atom_iconthemes = gdk_atom_intern ("_GTK_LOAD_ICONTHEMES", FALSE);
+
+	  for (i = 0; i < 5; i++)
+	    event->client.data.l[i] = 0;
+	  event->client.data_format = 32;
+	  event->client.message_type = atom_iconthemes;
+
+	  gdk_screen_broadcast_client_message (priv->screen, event);
+	}
+    }
 }
 
 /**
@@ -1181,6 +1202,8 @@ gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
 
   priv = icon_theme->priv;
 
+  GTK_NOTE (ICONTHEME, 
+	    g_print ("gtk_icon_theme_lookup_icon %s\n", icon_name));
   if (flags & GTK_ICON_LOOKUP_NO_SVG)
     allow_svg = FALSE;
   else if (flags & GTK_ICON_LOOKUP_FORCE_SVG)
@@ -1270,7 +1293,7 @@ gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
 			   "was not found either, perhaps you need to install it.\n"
 			   "You can get a copy from:\n"
 			   "\t%s"),
-			 icon_name, DEFAULT_THEME_NAME, "http://freedesktop.org/Software/icon-theme/releases");
+			 icon_name, DEFAULT_THEME_NAME, "http://icon-theme.freedesktop.org/releases");
 	    }
 	}
     }
@@ -1302,6 +1325,14 @@ gtk_icon_theme_error_quark (void)
  * and renders it into a pixbuf. This is a convenience function;
  * if more details about the icon are needed, use
  * gtk_icon_theme_lookup_icon() followed by gtk_icon_info_load_icon().
+ *
+ * Note that you probably want to listen for icon theme changes and
+ * update the icon. This is usually done by connecting to the 
+ * GtkWidget::style-set signal. If for some reason you do not want to
+ * update the icon when the icon theme changes, you should consider
+ * using gdk_pixbuf_copy() to make a private copy of the pixbuf
+ * returned by this function. Otherwise GTK+ may need to keep the old 
+ * icon theme loaded, which would be a waste of memory.
  * 
  * Return value: the rendered icon; this may be a newly created icon
  *  or a new reference to an internal icon, so you must not modify
@@ -1775,6 +1806,9 @@ theme_dir_get_icon_suffix (IconThemeDir *dir,
   else
       suffix = GPOINTER_TO_UINT (g_hash_table_lookup (dir->icons, icon_name));
 
+  GTK_NOTE (ICONTHEME, 
+	    g_print ("get_icon_suffix%s %d\n", dir->cache ? " (cached)" : "", suffix));
+
   return suffix;
 }
 
@@ -1815,6 +1849,8 @@ theme_lookup_icon (IconTheme          *theme,
     {
       dir = l->data;
 
+      GTK_NOTE (ICONTHEME, 
+		g_print ("theme_lookup_icon dir %s\n", dir->dir));
       suffix = theme_dir_get_icon_suffix (dir, icon_name, NULL);
       if (best_suffix (suffix, allow_svg) != ICON_SUFFIX_NONE)
 	{
@@ -1873,11 +1909,21 @@ theme_lookup_icon (IconTheme          *theme,
       
       if (min_dir->icon_data != NULL)
 	icon_info->data = g_hash_table_lookup (min_dir->icon_data, icon_name);
-      else if (min_dir->cache != NULL)
-	icon_info->data = _gtk_icon_cache_get_icon_data (min_dir->cache, icon_name, min_dir->subdir);
 
-      if (icon_info->data == NULL &&
-	  min_dir->cache && has_icon_file)
+      if (icon_info->data == NULL && min_dir->cache != NULL)
+	{
+	  icon_info->data = _gtk_icon_cache_get_icon_data (min_dir->cache, icon_name, min_dir->subdir);
+	  if (icon_info->data)
+	    {
+	      if (min_dir->icon_data == NULL)
+		min_dir->icon_data = g_hash_table_new_full (g_str_hash, g_str_equal,
+							    g_free, (GDestroyNotify)icon_data_free);
+
+	      g_hash_table_replace (min_dir->icon_data, g_strdup (icon_name), icon_info->data);
+	    }
+	}
+
+      if (icon_info->data == NULL && has_icon_file)
 	{
 	  gchar *icon_file_name, *icon_file_path;
 
@@ -1886,7 +1932,7 @@ theme_lookup_icon (IconTheme          *theme,
 
 	  if (g_file_test (icon_file_path, G_FILE_TEST_IS_REGULAR))
 	    {
-	      if (min_dir->icon_data == NULL)
+	      if (min_dir->icon_data == NULL)	
 		min_dir->icon_data = g_hash_table_new_full (g_str_hash, g_str_equal,
 							    g_free, (GDestroyNotify)icon_data_free);
 	      load_icon_data (min_dir, icon_file_path, icon_file_name);
@@ -2308,7 +2354,7 @@ gtk_icon_info_free (GtkIconInfo *icon_info)
     g_object_unref (icon_info->pixbuf);
   if (icon_info->cache_pixbuf)
     g_object_unref (icon_info->cache_pixbuf);
-  
+
   g_free (icon_info);
 }
 
@@ -2912,6 +2958,29 @@ find_builtin_icon (const gchar *icon_name,
   return min_icon;
 }
 
+void
+_gtk_icon_theme_check_reload (GdkDisplay *display)
+{
+  gint n_screens, i;
+  GdkScreen *screen;
+  GtkIconTheme *icon_theme;
+
+  n_screens = gdk_display_get_n_screens (display);
+  
+  for (i = 0; i < n_screens; i++)
+    {
+      screen = gdk_display_get_screen (display, i);
+
+      icon_theme = g_object_get_data (G_OBJECT (screen), "gtk-icon-theme");
+      if (icon_theme)
+	{
+	  icon_theme->priv->check_reload = TRUE;
+	  ensure_valid_themes (icon_theme);
+	  icon_theme->priv->check_reload = FALSE;
+	}
+    }
+}
+
 #ifdef G_OS_WIN32
 
 /* DLL ABI stability backward compatibility versions */
